
#include <windows.h>
#include "exports.h"
#include "cwstart.h"

LIBMAIN


int   DLL CobComHandler         ( long,  long  );
int   DLL CobComBufSaveHandler  ( long,  long  );
int   DLL CobComInitHandler     ( long,  long  );
LPSTR DLL CobComSetString       ( LPSTR        );
void  DLL CobComSetStringPrompt ( void         );
BOOL  DLL CobComPromptMe        ( BOOL         );
int   DLL CobComBegin           ( LPSTR, DWORD );
int   DLL CobComEnd             ( void         );
DWORD DLL CobComReset           ( void         );
DWORD DLL CobComQChanged        ( void         );
LPSTR DLL CobComQString         ( void         );
int   DLL CobComQStatus         ( void         );


void DLL 
_init()
{
	// Event handlers
	LibExport( "int   CobComHandler         LONG  LONG  ");
	LibExport( "int   CobComBufSaveHandler  LONG  LONG  ");
	LibExport( "int   CobComInitHandler     LONG  LONG  ");
	
	// Configuration routines
	LibExport( "LPSTR CobComSetString       LPSTR       ");
	LibExport( "void  CobComSetStringPrompt             ");
	LibExport( "BOOL  CobComPromptMe        BOOL        ");
	
	// Activate/deactivate commands
	LibExport( "int   CobComBegin           LPSTR DWORD ");
	LibExport( "int   CobComEnd                         ");
	LibExport( "DWORD CobComReset                       ");
	
	// Status query functions
	LibExport( "DWORD CobComQChanged                    ");
	LibExport( "LPSTR CobComQString                     ");
	LibExport( "int   CobComQStatus                     ");
}


/***************************
** Data global to the DLL **
***************************/

/*
 * Number of event handlers to install.
 * Must be less than or equal to 8, because QStatus uses only the bottom
 * two bytes as a bitmap of active handlers.
 * Handler 0 is reserved for the initialization handler.
 */
 
#define	EHANDLEMAX	(4)

HEVENT eHandle[EHANDLEMAX] = {0,0,0,0};

// This allocated string is inserted at the ends of modified lines.
LPSTR  cobComString        = NULL;

// This is the name of the file from which to read the comment string.
// It is _not_ an allocated string.  Do not modify nor free it.
LPSTR  cobComFile          = NULL;

// This flag is set or cleared by the like-named public function.
// If set, the user gets prompted for the comment string to use.
BOOL   cobComPromptMe      = FALSE;

// This is the column at which to insert the text.
DWORD  cobComColumn        = 73;

// Set by InitHandler; used only by CobComQStatus
static BOOL cobComInitialized = FALSE;

#define MYATTR_MODIFIED	(2)
#define MYATTR_INSERTED	(4)


/********************
** 	Event Handlers **
********************/

/*
 * Handler for line-changed events (lines changed or added).
 * First argument is the event that triggered this call.
 * Second argument is a pointer to a long int containing
 * the line number for which the event was generated.
 * Return value is 0, to indicate that processing should
 * continue normally.
 *
 * Line deletions are not handled.
 */
 
int DLL
CobComHandler(long event, long data)
{
	long  curLine; 
	LPSTR fileName;
	int   attr;    
	
	if (BufQSysFlags() & BUFFER_SYSTEM)
		return(0);

	if (!(fileName = BufQFilename()))
		return(0);
	
	if (!StrMatch(".*\\.cob", fileName, (int)(SEARCH_IGCASE | SEARCH_REGEX), NULL))
		if (!StrMatch(".*\\.cbl", fileName, (int)(SEARCH_IGCASE | SEARCH_REGEX), NULL))
			return(0);
	
	switch(event) 
	{
		case EVENT_LINES_CHANGED:
			if (data && data != -1 && *(long *)data)
			{
				curLine = *(long *)data;
				attr = AttrQVisible(); 
				if (!(attr & MYATTR_MODIFIED) && cobComColumn > 0)
				{
					int c_cur = BufReadChar();
					int c;

					MarkSavePos();
					MovAbs(curLine, cobComColumn - 1);
					c = BufReadChar();
					if (c != ' ' && c != '\t' && c != -1 && c != 0x0a)
					{
						long col = BufQCurrentCol();

						SrchFind("[ \t][^ \t]", SEARCH_REGEX, NULL);
						if (BufQCurrentCol() < col)
							MovAbs(curLine, cobComColumn);
					}
					else if ((c == ' ' || c == '\t') && c_cur == 0x0a)
						MovAbs(curLine, cobComColumn);
					BufDelToEOL();
					MarkRestorePos();
				}
				AttrSetVisible(curLine, curLine, attr | MYATTR_MODIFIED);
			}
			break;
		case EVENT_LINES_INSERTED:
			if (data && data != -1 && *(long *)data)
			{
				curLine = *(long *)data;
				attr = AttrQVisible();
				AttrSetVisible(curLine, curLine, attr | MYATTR_INSERTED);
			}
			break;
	}
			
	return(0);
}

/*
 * This handler is triggered on EVENT_BUFFER_CURRENT.  It de-installs
 * itself the first time it's called, after performing any needed
 * interaction with the user.  This setup ensures that the startup
 * process is complete before the user gets prompted.
 */

int DLL
CobComInitHandler(long event, long data)
{
	HBUFFER buffer = 0;
 	
	if (cobComFile && *cobComFile)
		if (buffer = BufEditFile(cobComFile))
		{
			MovAbs(1, 1);
			StrFree(cobComString);	// just in case an earlier call set this
			cobComString = BufReadStr(0x7FFFFFFF);
			BufDelBuffer(buffer);
		}	
	
	if (eHandle[0])
		if (!EventDeregister(eHandle[0]))
			eHandle[0] = 0;

	if (cobComPromptMe)
		CobComSetStringPrompt();

	// this variable is used only by CobComQStatus().
	cobComInitialized = TRUE;
	
	return(0);
}

/*
 * The handler called when a buffer is saved (before the buffer
 * is written to disk).
 * First argument is the event ID.
 * Second argument is NULL unless the buffer is being saved
 * under a name different from that of the original file.
 * It isn't used here anyway.
 * Return value is 0 to indicate that the buffer should be
 * saved normally.
 *
 * Looks at every line in the file; any that have been changed
 * are truncated at cobComColumn, and the template in the data file
 * is appended to the line at that column.
 */
 
int DLL
CobComBufSaveHandler(long event, long data)
{
	DWORD curLine;
	int   attr;   
	
	MarkSavePos();
	
	MovToLine(curLine = 1);
	
	while (!BufQPastEOF())
	{
		attr = AttrQVisible();
		if ((attr & MYATTR_INSERTED) || (attr & MYATTR_MODIFIED))
		{
			MovAbs(curLine, cobComColumn);
			BufDelToEOL();
			ExtExpandTemplate(cobComString);
		}	
		if (MovDown(1) != 1)
			break;
		curLine++;
	}
	MarkRestorePos();
	CobComReset();
	return(0);
}

/****************************
** 	Configuration routines **
****************************/
	
BOOL DLL
CobComPromptMe(BOOL yes_no)
{
	BOOL oldVal = cobComPromptMe;
	
	cobComPromptMe = yes_no;
	
	return(oldVal);
	
}
	
LPSTR DLL
CobComSetString(LPSTR newStr)
{
	if (newStr && *newStr)
	{
		StrFree(cobComString);
		cobComString = StrNew(newStr);
	}
	return(cobComString);
}

void DLL
CobComSetStringPrompt(void)
{
	CobComSetString(MsgPromptHistory("COBCOM", "COBOL Comment string: ",
	cobComString?cobComString:""));
}

/**********************************
** 	Activate/deactivate commands **
**********************************/
	
/* Install the line-changed and save-buffer handlers.
 * Return a bitmap of the handlers installed.
 *
 * The first argument sets the file containing the string to insert.
 * Whatever's at the beginning of the file will be used.
 * there is no checking for length or valid characters.
 * 
 * The string is global within this module.  It is
 * allocated here and is to be freed by CobComEnd.
 *
 * The string may contain Codewright macros.  Its only
 * use is that it is passed to ExtExpandTemplate, so
 * any legal argument for that function may be used here.
 *
 * The second argument is the column at which to insert
 * the string.  It defaults to 73.
 * 
 */
int DLL
CobComBegin(LPSTR fnam, DWORD col)
{
	int retval = 0;

	if (ArgCount >= 0)	// If we're called from .INI or LibFunctionExec, etc.
	{
		if (ArgCount() < 2)
			col = 0;
		if (ArgCount() < 1)
			fnam = NULL;
	}
	
	cobComFile = StrNew(fnam);

	if (col)
		cobComColumn = col;

	if (fnam && *fnam)
		if (!eHandle[0])
			if (eHandle[0] = EventRegister(EVENT_BUFFER_CURRENT, EVENT_NORMAL,
			"CobComInitHandler"))
				retval |= 1;
			
	if (!eHandle[1])
		if (eHandle[1] = EventRegister(EVENT_LINES_INSERTED, EVENT_NORMAL,
		"CobComHandler"))
			retval |= 2;
			
	if (!eHandle[2])
		if (eHandle[2] = EventRegister(EVENT_LINES_CHANGED, EVENT_NORMAL,
		"CobComHandler"))
			retval |= 4;

	if (!eHandle[3])
		if (eHandle[3] = EventRegister(EVENT_SAVE_BUFFER, EVENT_NORMAL,
		"CobComBufSaveHandler"))
			retval |= 8;

	return(retval);
}

/* Turn off all installed handlers.
 * This call removes any effect of the DLL.
 * A bitmap of the successfully deinstalled handlers
 * is returned.
 */
 
int DLL
CobComEnd(void)
{
	BOOL retval = FALSE;
	int  i;    

	for (i = 0; i < EHANDLEMAX; ++i)
	{
		// Deregister returns 0 if OK
		if (eHandle[i] && !EventDeregister(eHandle[i]))
		{
			eHandle[i] = 0;
			retval |= 1 << i;
		}
	}
	if (i >= EHANDLEMAX)
	{
		StrFree(cobComString);
		cobComString = NULL;
		StrFree(cobComFile);  
		cobComFile = NULL;
	}
		
	return(retval);
}

/*
 * This function is not called automatically.  It may
 * be called from an API command prompt.
 *
 * It clears the changed status of each line in the file.
 * If you call this, only lines changed after the call
 * will get tagged.
 *
 * Return value is the number of lines that had been
 * targetted for tagging.
 */
 
DWORD DLL
CobComReset(void)
{
	DWORD resetCount = 0;
	DWORD curLine;  
	int   attr;     
	
	MarkSavePos();
	
	MovToLine(curLine = 1);
	
	while (!BufQPastEOF())
	{
		attr = AttrQVisible();
		if (attr & (MYATTR_INSERTED | MYATTR_MODIFIED))
		{
			attr &= ~(MYATTR_INSERTED | MYATTR_MODIFIED);
			AttrSetVisible(curLine, curLine, attr);
			resetCount++;
		}
		if (MovDown(1) != 1)
			break;
		curLine++;
	}
	MarkRestorePos();
	return(resetCount);
}

/****************************
** 	Status query functions **
****************************/

/* Diagnostic tool: returns the string that is to be inserted
 * into the comment field.  String must not be deallocated by
 * the caller.
 */

LPSTR DLL
CobComQString(void)
{
	if (cobComString && *cobComString)
		return(cobComString);
	return(NULL);	
}

/* Diagnostic tool: tells which handlers are installed
 * (handler 0 is the self-deinstalling init handler).
 * Returns a bitmap of the handlers installed, along with
 * bits to indicate the state of DLL-global variables.
 */

int DLL
CobComQStatus(void)
{
	int i;
	int retval = 0;
	
	for (i = 0; i < EHANDLEMAX; ++i)
	{
		if (eHandle[i])
			retval |= 1 << i;
	}
	if (cobComInitialized)
		retval |= 0x100;
	if (cobComFile && *cobComFile)	
		retval |= 0x200;
	if (cobComString && *cobComString)	
		retval |= 0x400;
	return(retval);	
}

	
/*
 * This function returns the number of lines that have been
 * targeted for tagging.  It also notifies the user of each individual
 * targetted line.  It is a diagnostic tool.
 */	
DWORD DLL
CobComQChanged(void)
{
	DWORD addedCount    = 0;
	DWORD modifiedCount = 0;
	DWORD curLine;     
	int   attr;        
	
	MarkSavePos();
	
	MovToLine(curLine = 1);
	
	while (!BufQPastEOF())
	{
		attr = AttrQVisible();
		if (attr & MYATTR_INSERTED)
		{
			MsgPrintf(3, "New line at %ld", curLine);
			addedCount++;
		}	
		if (attr & MYATTR_MODIFIED)
		{
			MsgPrintf(3, "Changed line at %ld", curLine);
			modifiedCount++;
		}	
		if (MovDown(1) != 1)
			break;
		curLine++;
	}
	MarkRestorePos();
	MsgPrintf(3, "SUMMARY: %ld line%s added", addedCount, (addedCount==1)?"":"s");	
	MsgPrintf(3, "SUMMARY: %ld line%s changed", modifiedCount, (modifiedCount==1)?"":"s");
	return(addedCount + modifiedCount);
}
			
